# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

"""Provides a GuidBasedRecommender for use in production

Adds in the S3 context with the help of the srgutil Context.
"""

import numpy as np

from decouple import config
from srgutil.interfaces import IS3Data, IMozLogging
from .utils.cache import LazyJSONLoader
from .recommenders.guidguid import GuidGuidCoinstallRecommender
from .recommenders.treatments import (
    Guidception,
    NoTreatment,
    RowCount,
    RowNormSum,
    RowSum
)

ADDON_LIST_BUCKET = 'telemetry-parquet'
ADDON_LIST_KEY = 'taar/lite/guid_coinstallation.json'
GUID_RANKING_KEY = 'taar/lite/guid_install_ranking.json'

ADDON_DL_ERR = "Cannot download addon coinstallation file {}".format(ADDON_LIST_KEY)   # noqa
TAAR_CACHE_EXPIRY = config('TAAR_CACHE_EXPIRY', default=14400, cast=int)

NORM_MODE_ROWNORMSUM = 'rownorm_sum'
NORM_MODE_ROWCOUNT = 'row_count'
NORM_MODE_ROWSUM = 'row_sum'
NORM_MODE_GUIDCEPTION = 'guidception'

# TODO BirdNote - Delete before merging
# This only required modest changes to use the new
# Recommender and Treatment classes. The main difference is that there's
# no norm_dict instead there's the _recommenders dict which gets rebuilt
# whenever json is reloaded and then that is used directly to pull recommendations


class GuidBasedRecommender:
    """ A recommender class that returns top N addons based on a
    passed addon identifier.  This will load a json file containing
    updated top n addons coinstalled with the addon passed as an input
    parameter based on periodically updated  addon-addon
    coinstallation frequency table generated from  Longitdudinal
    Telemetry data.  This recommender will drive recommendations
    surfaced on addons.mozilla.org
    """

    _addons_coinstallations = None
    _recommenders = {}

    # Define recursion levels for guid-ception
    RECURSION_LEVELS = 3

    def __init__(self, ctx):
        self._ctx = ctx
        assert IS3Data in self._ctx

        if 'coinstall_loader' in self._ctx:
            self._addons_coinstall_loader = self._ctx['coinstall_loader']
        else:
            self._addons_coinstall_loader = LazyJSONLoader(self._ctx,
                                                           ADDON_LIST_BUCKET,
                                                           ADDON_LIST_KEY,
                                                           TAAR_CACHE_EXPIRY)

        if 'ranking_loader' in self._ctx:
            self._guid_ranking_loader = self._ctx['ranking_loader']
        else:
            self._guid_ranking_loader = LazyJSONLoader(self._ctx,
                                                       ADDON_LIST_BUCKET,
                                                       GUID_RANKING_KEY,
                                                       TAAR_CACHE_EXPIRY)

        self._init_from_ctx()

        # Force access to the JSON models for each request at
        # recommender construction.  This was lifted out of the
        # constructor for the LazyJSONLoader so that the
        # precomputation of the normalization tables can be done in
        # the recommender.
        _ = self._addons_coinstallations  # noqa
        _ = self._guid_rankings           # noqa

        self.logger.info("GUIDBasedRecommender is initialized")

    def _init_from_ctx(self):
        self.logger = self._ctx[IMozLogging].get_logger('taarlite')

        if self._addons_coinstallations is None:
            self.logger.error(ADDON_DL_ERR)

        # Compute the floor install incidence that recommended addons
        # must satisfy.  Take 5% of the mean of all installed addons.
        self._min_installs = np.mean(list(self._guid_rankings.values())) * 0.05

        # Warn if the minimum number of installs drops below 100.
        if self._min_installs < 100:
            self.logger.warn("minimum installs threshold low: [%s]" % self._min_installs)

    @property
    def _addons_coinstallations(self):
        result, refreshed = self._addons_coinstall_loader.get()
        if refreshed:
            self.logger.info("Refreshing guid_maps for normalization")
            self._precompute_recommenders()
        return result

    @property
    def _guid_rankings(self):
        result, refreshed = self._guid_ranking_loader.get()
        if refreshed:
            self.logger.info("Refreshing guid_maps for normalization")
            self._precompute_recommenders()
        return result

    def _precompute_recommenders(self):
        def get_recommender(treatment):
            return GuidGuidCoinstallRecommender(
                self._addons_coinstallations,
                self._guid_rankings,
                treatment,
                validate_raw_coinstallation_graph=False
            )
        self._recommenders = {
            'none': get_recommender(NoTreatment()),
            NORM_MODE_ROWCOUNT: get_recommender(RowCount()),
            NORM_MODE_ROWSUM: get_recommender(RowSum()),
            NORM_MODE_ROWNORMSUM: get_recommender(RowNormSum()),
            NORM_MODE_GUIDCEPTION: get_recommender(Guidception()),
        }

    def recommend(self, client_data, limit=4):
        """
        TAAR lite will yield 4 recommendations for the AMO page
        """

        # Force access to the JSON models for each request at the
        # start of the request to update normalization tables if
        # required.
        _ = self._addons_coinstallations  # noqa
        _ = self._guid_rankings           # noqa

        addon_guid = client_data.get('guid')
        normalize = client_data.get('normalize', NORM_MODE_ROWNORMSUM)

        if normalize is not None and normalize not in self._recommenders:
            # Yield no results if the normalization method is not specified
            self.logger.warn("Invalid normalization parameter detected: [%s]" % normalize)
            return []

        result_list = self._recommenders[normalize].recommend(addon_guid, limit)

        log_data = (str(addon_guid), [str(r) for r in result_list])
        self.logger.info("Addon: [%s] triggered these recommendation guids: [%s]" % log_data)

        return result_list

    def can_recommend(self, client_data):
        # TODO - I can't see anywhere this is used.

        # We can't recommend if we don't have our data files.
        if self._addons_coinstallations is None:
            return False

        # If we have data coming from other sources, we can use that for
        # recommending.
        addon_guid = client_data.get('guid', None)
        if not isinstance(addon_guid, str):
            return False

        # Use a dictionary keyed on the query guid
        if addon_guid not in self._addons_coinstallations.keys():
            return False

        if not self._addons_coinstallations.get(addon_guid):
            return False

        return True



import json
import boto3
from moto import mock_s3
from taar_lite.production import (
    GuidBasedRecommender,
    ADDON_LIST_BUCKET,
    ADDON_LIST_KEY,
    GUID_RANKING_KEY,
)
from taar_lite.utils.cache import LazyJSONLoader


@mock_s3
def test_logging(default_ctx, MOCK_DATA, MOCK_GUID_RANKING):
    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=ADDON_LIST_BUCKET)
    conn.Object(ADDON_LIST_BUCKET, ADDON_LIST_KEY)\
        .put(Body=json.dumps(MOCK_DATA))
    conn.Object(ADDON_LIST_BUCKET, GUID_RANKING_KEY)\
        .put(Body=json.dumps(MOCK_GUID_RANKING))
    coinstall_loader = LazyJSONLoader(default_ctx,
                                      ADDON_LIST_BUCKET,
                                      ADDON_LIST_KEY)

    ranking_loader = LazyJSONLoader(default_ctx,
                                    ADDON_LIST_BUCKET,
                                    GUID_RANKING_KEY)

    default_ctx['coinstall_loader'] = coinstall_loader
    default_ctx['ranking_loader'] = ranking_loader

    recommender = GuidBasedRecommender(default_ctx)

    # These would error out if the object type was incorrect
    recommender.logger.error('foo')
    recommender.logger.warn('bar')
    recommender.logger.info('foo')
    recommender.logger.debug('bar')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from .treatments import BaseTreatment


class GuidGuidCoinstallRecommender:
    """ A recommender class that returns top N addons based on a
    passed addon identifier.
    Accepts:
        - a dict containing coinstalled addons
        - a dict of addon rankins
        - a instance of a treatment class that transforms the original coinstallation dict

    Provides a recommend method to then return recommendations for a supplied addon.
    Can also return the complete recommendation graph.
    """

    def __init__(
            self,
            raw_coinstallation_graph,
            guid_rankings,
            treatment_cls,
            apply_treatment_on_init=True,
            validate_raw_coinstallation_graph=True):
        assert isinstance(treatment_cls, BaseTreatment)
        if validate_raw_coinstallation_graph:
            self.validate_raw_coinstallation_graph(raw_coinstallation_graph)
        self._treatment_cls = treatment_cls
        self._raw_coinstallation_graph = raw_coinstallation_graph
        self._guid_rankings = guid_rankings
        self._treatment_graph = {}
        self._min_installs = 0
        if apply_treatment_on_init:
            self.build_treatment_graph()

    @classmethod
    def validate_raw_coinstallation_graph(cls, coinstallations):
        # I have a recollection of problems importing pandas in
        # production, so I've wrapped the imports here and
        # currently validation is set off for production by default.
        import numpy as np
        import pandas as pd

        sorted_guids = sorted(list(coinstallations.keys()))
        df = pd.DataFrame(coinstallations, index=sorted_guids, columns=sorted_guids)
        as_matrix = df.values
        assert np.allclose(as_matrix, as_matrix.T, equal_nan=True)

    @property
    def min_installs(self):
        """Returns the minimum number of installs acceptable to keep a guid in the recommendations."""
        return self._min_installs

    @property
    def raw_coinstallation_graph(self):
        """Returns a dictionary with guid keys and a coinstall set values.

        Something like this, but with much longer values.

            {
                'guid_a': {'guid_b': 10, 'guid_c': 13},
                'guid_b': {'guid_a': 10, 'guid_c': 4},
                'guid_c': {'guid_a': 13, 'guid_b': 4}
            }

        It must be symmetric.
        """
        return self._raw_coinstallation_graph

    @property
    def treatment_graph(self):
        """Returns the recommentaion graph.

        Recommendation graph is in the same format as the coinstallation graph but the
        numerical values are the weightings as a result of the treatment.
        """
        return self._treatment_graph

    @property
    def guid_rankings(self):
        """Returns a dictionary with guid keys and install count values"""
        return self._guid_rankings

    def get_recommendation_graph(self, limit):
        """The recommendation graph is the full output for all addons"""
        rec_graph = {}
        for guid in self.raw_coinstallation_graph:
            rec_graph[guid] = self.recommend(guid, limit)
        return rec_graph

    def build_treatment_graph(self):
        """Does the work to compute and then set the recommendation graph.
        Sub classes may wish to override if more complex computation is required.
        """
        # TODO Is this too coupled? (Am okay leaving for now)
        self._treatment_graph = self._treatment_cls.treat(self.raw_coinstallation_graph)

    def recommend(self, for_guid, limit):
        """Returns a list of sorted recommendations of length 0 - limit for supplied guid.

        Result list is a list of tuples with the lex ranking string. e.g.
            [
                ('guid_a', '000003.000002.0001000'),
                ('guid_c', '000001.000002.0001000'),
                ('guid_b', '000001.000002.0000010'),
            ]

        """
        if for_guid not in self.treatment_graph:
            print('here')
            return []
        raw_recommendations = self.treatment_graph[for_guid]
        cleaned_recommendations = self._strip_low_ranked_guids(raw_recommendations)
        result_list = self._build_sorted_result_list(cleaned_recommendations)
        return result_list[:limit]

    def _strip_low_ranked_guids(self, input_dict):
        """Takes a dictionary with a format matching the values in the coinstall_dict
        and strips it of keys that do not meet the minimum installs.

            In:  {'guid_b': 10, 'guid_c': 13}
            Out: {'guid_b': 10, 'guid_c': 13}
        """

        cleaned_dict = {}
        for k, v in input_dict.items():
            if self.guid_rankings.get(k, 0) >= self.min_installs:
                cleaned_dict[k] = v
        return cleaned_dict

    def _build_sorted_result_list(self, unranked_recommendations):
        """Takes a dictionary with a format matching the values in the coinstall_dict
        and return a sorted list of results

            In: {'guid_b': 10, 'guid_c': 13}
            Out:
                [
                    ('guid_c', '000013.000000.0001000'),
                    ('guid_b', '000010.000000.0000010'),
                ]

        """
        # Augment the result_dict with the installation counts
        # and then we can sort using lexical sorting of strings.
        # The idea here is to get something in the form of
        #    0000.0000.0000
        # The computed weight takes the first and second segments of
        # integers.  The third segment is the installation count of
        # the addon but is zero padded.
        result_dict = {}
        for k, v in unranked_recommendations.items():
            lex_value = "{0:020.10f}.{1:010d}".format(v, self.guid_rankings.get(k, 0))
            result_dict[k] = lex_value
        # Sort the result dictionary in descending order by weight
        result_list = sorted(result_dict.items(), key=lambda x: x[1], reverse=True)
        return result_list



class BaseTreatment:

    def treat(self, raw_coinstallation_dict):
        """Accept a coinstallation graph, and returns a treated graph.
        No constraints are put on the shape of the return graph but the format
        is expected to be the same.

        Graph format:

            {
                'guid_a': {'guid_b': 10, 'guid_c': 13},
                'guid_b': {'guid_a': 10, 'guid_c': 4},
                'guid_c': {'guid_a': 13, 'guid_b': 4}
            }

        """
        raise NotImplementedError


class NoTreatment(BaseTreatment):
    """Returns the original coinstallation dict"""
    def treat(self, input_coinstall_dict):
        return input_coinstall_dict


class RowSum(BaseTreatment):
    """This normalization normalizes the weights for the suggested
    coinstallation GUIDs based on the sum of the weights for the
    coinstallation GUIDs.
    """
    def treat(self, input_coinstall_dict):
        guid_count_map = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            for coinstall_guid, coinstall_count in coinstalls.items():
                guid_count_map.setdefault(coinstall_guid, 0)
                guid_count_map[coinstall_guid] += coinstall_count

        treatment_dict = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            output_dict = {}
            for guid, guid_weight in coinstalls.items():
                norm_guid_weight = guid_weight * 1.0 / guid_count_map[guid]
                output_dict[guid] = norm_guid_weight
            treatment_dict[guidkey] = output_dict

        return treatment_dict


class RowCount(BaseTreatment):
    """This normalization method counts the unique times that a
    GUID is coinstalled with any other GUID.

    This dampens weight of any suggested GUID inversely
    proportional to it's overall popularity.
    """

    def treat(self, input_coinstall_dict):
        row_count = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            for coinstall_guid, _ in coinstalls.items():
                row_count.setdefault(coinstall_guid, 0)
                row_count[coinstall_guid] += 1

        treatment_dict = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            output_dict = {}
            for result_guid, result_count in coinstalls.items():
                output_dict[result_guid] = 1.0 * result_count / row_count[result_guid]
            treatment_dict[guidkey] = output_dict

        return treatment_dict


class RowNormalizationMixin():

    def _normalize_row_weights(self, coinstall_dict):
        # Compute an intermediary dictionary that is a row normalized
        # co-install. That is - each coinstalled guid weight is
        # divided by the sum of the weights for all coinstalled guids
        # on this row.
        tmp_dict = {}
        coinstall_total_weight = sum(coinstall_dict.values())
        for coinstall_guid, coinstall_weight in coinstall_dict.items():
            tmp_dict[coinstall_guid] = coinstall_weight / coinstall_total_weight
        return tmp_dict


class RowNormSum(BaseTreatment, RowNormalizationMixin):
    """This normalization is the same as norm_row_sum, but we also
    divide the result by the sum of
    (addon coinstall instances)/(addon coinstall total instances)

    The testcase for this scenario lays out the math more
    explicitly.
    """

    def _build_guid_row_norm(self, input_coinstall_dict):
        guid_row_norm = {}
        for _, coinstalls in input_coinstall_dict.items():
            rowsum = sum(coinstalls.values())
            for coinstall_guid, coinstall_count in coinstalls.items():
                if coinstall_guid not in guid_row_norm:
                    guid_row_norm[coinstall_guid] = []
                guid_row_norm[coinstall_guid].append(1.0 * coinstall_count / rowsum)
        return guid_row_norm

    def treat(self, input_coinstall_dict):
        guid_row_norm = self._build_guid_row_norm(input_coinstall_dict)
        treatment_dict = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            output_dict = {}
            tmp_dict = self._normalize_row_weights(coinstalls)
            for output_guid, output_guid_weight in tmp_dict.items():
                guid_row_norm_list = guid_row_norm.get(output_guid, [])
                norm_sum = sum(guid_row_norm_list)
                output_dict[output_guid] = output_guid_weight / norm_sum
            treatment_dict[guidkey] = output_dict

        return treatment_dict


class Guidception(BaseTreatment, RowNormalizationMixin):

    # Define recursion levels for guid-ception
    RECURSION_LEVELS = 3
    _coinstallations = {}

    def treat(self, input_coinstall_dict):
        self._coinstallations = input_coinstall_dict
        treatment_dict = {}
        for guidkey, coinstalls in input_coinstall_dict.items():
            tmp_dict = self._normalize_row_weights(coinstalls)
            output_dict = self._compute_recursive_results(tmp_dict, self.RECURSION_LEVELS)
            treatment_dict[guidkey] = output_dict
        return treatment_dict

    def _recursion_penalty(self, level):
        """ Return a factor to apply to the weight for a guid recommendation."""
        dampener = 1.0 - (1.0 * (self.RECURSION_LEVELS - level) / self.RECURSION_LEVELS)
        dampener *= dampener
        return dampener

    def _compute_recursive_results(self, row_normalized_coinstall, level):
        if level <= 0:
            return row_normalized_coinstall

        # consolidated_coinstall_dict will capture values
        consolidated_coinstall_dict = {}

        # Add this level's guid weight to the consolidated result
        dampener = self._recursion_penalty(level)
        for _, _ in row_normalized_coinstall.items():
            for guid, guid_weight in row_normalized_coinstall.items():
                weight = consolidated_coinstall_dict.get(guid, 0)
                weight += (dampener*guid_weight)
                consolidated_coinstall_dict[guid] = weight

        # Add in the next level
        level -= 1
        for guid in consolidated_coinstall_dict:
            next_level_coinstalls = self._coinstallations.get(guid, {})
            if next_level_coinstalls != {}:
                # Normalize the next bunch of suggestions
                next_level_coinstalls = self._normalize_row_weights(next_level_coinstalls)

                next_level_results = self._compute_recursive_results(next_level_coinstalls, level)
                for _, next_level_weight in next_level_results.items():
                    weight = consolidated_coinstall_dict.get(guid, 0)
                    weight += next_level_weight
                    consolidated_coinstall_dict[guid] = weight

        # normalize the final results
        return self._normalize_row_weights(consolidated_coinstall_dict)
